import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# plt.style.use("dark_background")

files = [
    "straight.csv",
    "10deg.csv",
    "20deg.csv",
    "30deg.csv",
    "40deg.csv",
    "50deg.csv",
    "60deg.csv",
]
angles = {
    "straight.csv": 0,
    "10deg.csv": 10,
    "20deg.csv": 20,
    "30deg.csv": 30,
    "40deg.csv": 40,
    "50deg.csv": 50,
    "60deg.csv": 60,
}
c = 343
rho = 1.21
d = 40e-3
r = 7.62e-2 / 2


def load_data(name):
    """
    Load data from a csv file
    """
    return pd.read_csv(name, encoding="unicode_escape")


def data_to_numpy(data):
    """
    Convert data to numpy arrays
    """
    f = data[" Frequency (Hz)"].to_numpy()
    G = data[" Gain (dB)"].to_numpy()
    P = data[" Phase (°)"].to_numpy()
    return f, G, P


def transfer_function(G, P):
    """
    Convert data to transfer function
    """
    H = 10 ** (G / 20) * (np.cos(np.radians(P)) + 1j * np.sin(np.radians(P)))
    return H


def wave_number(f):
    """
    Calculate wave number
    """
    return 2 * np.pi * f / c


def reflection_coefficient(H, k):
    """
    Calculate reflection coefficient
    """
    return (H - np.exp(-1j * k * d)) / (np.exp(-1j * k * d) - H)


def impedance(reflection):
    """
    Calculate impedance
    """
    Z = rho * c * (1 + reflection) / (1 - reflection)
    return Z / (np.pi * r**2) / 1000


def plot_data(data):
    """
    Plot impedance data
    """
    plt.figure()
    for angle in data:
        f, Z = data[angle]
        color = plt.cm.viridis(angle / 60)
        plt.subplot(211)
        plt.plot(f, np.real(Z), label=f"{angle}°", color=color)
        plt.subplot(212)
        plt.plot(f, np.imag(Z), label=f"{angle}°", color=color)
    plt.subplot(211)
    plt.ylabel("$R$ (Pa s/$\\mathrm{m}^3$)")
    plt.subplot(212)
    plt.ylabel("$X$ (Pa s/$\\mathrm{m}^3$)")
    plt.xlabel("$f$ (Hz)")
    # Show legend to the right of the plot outside the plot
    plt.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    plt.show()


def plot_polar(data):
    """
    Make polar subplots for resistance and reactance that show the each point in a color according to the magnitude of the real or complex component and the radius being the frequency.
    """
    fig, ax = plt.subplots(2, 1, subplot_kw=dict(projection="polar"))
    for angle in data:
        f, Z = data[angle]
        ax[0].scatter(np.angle(Z), f, c=np.abs(np.real(Z)), cmap="viridis")
        ax[1].scatter(np.angle(Z), f, c=np.abs(np.imag(Z)), cmap="viridis")
    ax[0].set_title("Resistance")
    ax[1].set_title("Reactance")
    # Force plot to be square
    fig.set_size_inches(6, 12)
    # Radial axis label as "Frequency (Hz)"
    ax[0].set_rlabel_position(135)
    # Add colorbar
    plt.colorbar(ax[0].collections[0], ax=ax[0], label="Magnitude")
    plt.show()


def plot_heatmap(data, filename=""):
    """
    Make a heatmap for resistance and reactance that shows the magnitude of the real or complex component as a function of the frequency and the angle.
    """
    cmap = "viridis"
    fig, ax = plt.subplots(2, 1)
    fig.suptitle("Acoustic Impedance vs Frequency and Angle")
    # Convert data into array where rows are angles and columns are frequencies
    R = np.array([np.real(data[angle][1]) for angle in data])
    I = np.array([np.imag(data[angle][1]) for angle in data])
    # Plot resistance
    im0 = ax[0].imshow(R, aspect="auto", cmap=cmap, extent=[10, 35, 0, 60], interpolation="gaussian")
    plt.colorbar(im0, ax=ax[0], label="Resistance (Pa s/L)")
    ax[0].set_xlabel("Frequency (Hz)")
    ax[0].set_ylabel("Angle (°)")

    # Plot reactance
    im1 = ax[1].imshow(I, aspect="auto", cmap=cmap, extent=[10, 35, 0, 60], interpolation="gaussian")
    plt.colorbar(im1, ax=ax[1], label="Reactance (Pa s/L)")
    ax[1].set_xlabel("Frequency (Hz)")
    ax[1].set_ylabel("Angle (°)")

    plt.show()

    if filename:
        fig.savefig(filename, dpi=300)


def main():
    data = {}

    for file in files:
        oscilloscope_data = load_data(file)
        f, G, P = data_to_numpy(oscilloscope_data)
        H = transfer_function(G, P)
        k = wave_number(f)
        reflection = reflection_coefficient(H, k)
        Z = impedance(reflection)

        data[angles[file]] = f, Z

    # plot_data(data)
    # plot_polar(data)
    plot_heatmap(data, filename="heatmap.png")


if __name__ == "__main__":
    main()
